"""
====================
File Path Management
====================

"""

from datetime import datetime
from fnmatch import fnmatch
from pathlib import Path
from typing import NamedTuple

from vivarium.interface.utilities import get_output_model_name_string

from vivarium_cluster_tools import utilities as vct_utils
from vivarium_cluster_tools.psimulate import COMMANDS

DEFAULT_LOAD_TESTS_DIR = "/mnt/team/simulation_science/priv/engineering/load_tests"
CENTRAL_PERFORMANCE_LOGS_DIRECTORY = Path(
    "/mnt/team/simulation_science/pub/performance_logs/"
)


class InputPaths(NamedTuple):
    """Container class for input filepaths.

    In addition to the path attributes, this class provides a method to create
    the required directories.

    """

    model_specification: Path | None
    """The path to the model specification file."""
    branch_configuration: Path | None
    """The path to the branch configuration file."""
    artifact: Path | None
    """The path to the data artifact."""
    result_directory: Path
    """The path to the results directory."""

    @classmethod
    def from_entry_point_args(
        cls,
        *,  # No positional args allowed.
        result_directory: str | Path,
        input_model_specification_path: str | Path | None = None,
        input_branch_configuration_path: str | Path | None = None,
        input_artifact_path: str | Path | None = None,
    ) -> "InputPaths":
        """Create an instance of InputPaths from the arguments passed to the entry point.

        Parameters
        ----------
        result_directory
            The path to the results directory.
        input_model_specification_path
            The path to the model specification file.
        input_branch_configuration_path
            The path to the branch configuration file.
        input_artifact_path
            The path to the data artifact.

        Returns
        -------
            An instance of InputPaths.

        """

        results_directory_path = cls._coerce_path(result_directory)
        if results_directory_path is None:
            raise ValueError("Result directory must be provided.")

        return InputPaths(
            model_specification=cls._coerce_path(input_model_specification_path),
            branch_configuration=cls._coerce_path(input_branch_configuration_path),
            artifact=cls._coerce_path(input_artifact_path),
            result_directory=results_directory_path,
        )

    @staticmethod
    def _coerce_path(path: str | Path | None) -> Path | None:
        if path is not None:
            return Path(path) if not isinstance(path, Path) else path
        return None


class OutputPaths(NamedTuple):
    """Container class for output filepaths.

    In addition to the path attributes, this class provides a method to create
    the required directories as well as a class method to create an instance
    from the arguments passed to the entry point.

    """

    # Directories
    root: Path
    """The timestamped root directory for the simulation output."""

    logging_root: Path
    """The parent directory for all logs."""
    cluster_logging_root: Path
    """The root directory for cluster logs."""
    worker_logging_root: Path
    """The root directory for worker logs."""

    # Files
    # Environment configuration
    worker_settings: Path
    """The path to the worker settings file."""
    environment_file: Path
    """The path to the requirements.txt environment file."""

    # Simulation configuration
    model_specification: Path
    """The path to the model specification file."""
    keyspace: Path
    """The path to the simulation keyspace file."""
    branches: Path
    """The path to the simulation branches file."""

    # outputs
    finished_sim_metadata: Path
    """The path to the finished simulation metadata file."""
    results_dir: Path
    """The path to the results directory."""
    backup_dir: Path
    """The path to the simulation backup directory."""
    backup_metadata_path: Path
    """The path to the backup metadata file."""

    # will not be reliable if we parallelized across artifacts
    @property
    def artifact_name(self) -> str:
        """Name of the artifact."""
        return self.root.parent.stem

    @property
    def run_date(self) -> str:
        """Date of the simulation run."""
        runtime_info = self.logging_root.stem
        run_date = runtime_info[: runtime_info.rindex("_")]
        return run_date

    @property
    def run_type(self) -> str:
        """Type of the simulation run."""
        runtime_info = self.logging_root.stem
        run_type = runtime_info[runtime_info.rindex("_") + 1 :]
        return run_type

    @property
    def original_run_date(self) -> str:
        """Date of the original simulation run."""
        return self.root.stem

    @property
    def project_name(self) -> str:
        """Name of the project."""
        if self.logging_to_central_results_directory:
            return self.root.parents[3].stem
        else:
            return self.root.parents[1].stem

    @property
    def root_path(self) -> str:
        """Path to the root directory."""
        return str(self.root.parent)

    @property
    def logging_to_central_results_directory(self) -> bool:
        """Whether the logs are being written to the central results directory."""
        return fnmatch(str(self.root), "/mnt/team/simulation_science/pub/models/*/results/*")

    @classmethod
    def from_entry_point_args(
        cls,
        *,  # No positional args allowed.
        command: str,
        input_artifact_path: Path | None,
        result_directory: Path,
        input_model_spec_path: Path | None,
    ) -> "OutputPaths":
        """Create an instance of OutputPaths from the arguments passed to the entry point.

        Parameters
        ----------
        command
            The specific ``psimulate`` command being run.
        input_artifact_path
            The path to the data artifact.
        result_directory
            The path to the results directory.
        input_model_spec_path
            The path to the model specification file.

        Returns
        -------
            An instance of OutputPaths.

        """
        launch_time = datetime.now().strftime("%Y_%m_%d_%H_%M_%S")

        output_directory = result_directory
        if command == COMMANDS.run:
            if input_model_spec_path is None:
                raise ValueError(
                    "Model specification path must be provided for 'run' command."
                )
            model_name = get_output_model_name_string(
                input_artifact_path, input_model_spec_path
            )
            output_directory = output_directory / model_name / launch_time
        elif command == COMMANDS.load_test:
            output_directory = output_directory / launch_time

        logging_directory = output_directory / "logs" / f"{launch_time}_{command}"
        logging_dirs = {
            "logging_root": logging_directory,
            "cluster_logging_root": logging_directory / "cluster_logs",
            "worker_logging_root": logging_directory / "worker_logs",
        }

        output_paths = OutputPaths(
            root=output_directory,
            **logging_dirs,
            worker_settings=output_directory / "settings.py",
            environment_file=output_directory / "requirements.txt",
            model_specification=output_directory / "model_specification.yaml",
            keyspace=output_directory / "keyspace.yaml",
            branches=output_directory / "branches.yaml",
            finished_sim_metadata=output_directory / "finished_sim_metadata.csv",
            results_dir=output_directory / "results",
            backup_dir=output_directory / "sim_backups",
            backup_metadata_path=output_directory / "sim_backups" / "backup_metadata.csv",
        )
        return output_paths

    def touch(self) -> None:
        """Create the required directories."""
        for dir in [self.root, self.results_dir, self.backup_dir]:
            vct_utils.mkdir(dir, exists_ok=True, parents=True)
        for dir in [self.logging_root, self.cluster_logging_root, self.worker_logging_root]:
            vct_utils.mkdir(dir, parents=True)



"""
=========
redis_dbs
=========

Redis database management.

"""
from vivarium_cluster_tools.psimulate.redis_dbs.cli_options import (
    with_max_workers,
    with_redis,
)
from vivarium_cluster_tools.psimulate.redis_dbs.launcher import (
    launch_redis_processes as launch,
)
from vivarium_cluster_tools.psimulate.redis_dbs.registry import RegistryManager



"""
================================
vivarium_cluster_tools Utilities
================================

Making directories is hard.

"""

import functools
import os
import socket
import time
import warnings
from collections.abc import Callable
from pathlib import Path
from typing import Any, ParamSpec, TypeVar

from vivarium_cluster_tools.psimulate.environment import ENV_VARIABLES

NUM_ROWS_PER_CENTRAL_LOG_FILE = 100_000

P = ParamSpec("P")
T = TypeVar("T")


def get_cluster_name() -> str:
    """Returns the hostname"""
    return socket.gethostname()


def get_drmaa() -> Any:
    try:
        import drmaa
    except (RuntimeError, OSError):
        if "slurm" in ENV_VARIABLES.HOSTNAME.value:
            ENV_VARIABLES.DRMAA_LIB_PATH.update("/opt/slurm-drmaa/lib/libdrmaa.so")
            import drmaa
        else:
            drmaa = object()
    return drmaa


def mkdir(
    path: str | Path, umask: int = 0o002, exists_ok: bool = False, parents: bool = False
) -> None:
    """Utility method to create a directory with specified permissions.

    Parameters
    ----------
    path
        Path of the directory to create.
    umask
        Umask specifying the desired permissions - defaults to 0o002.
    exists_ok
        If False, raises FileExistsError if the directory already exists.
    parents
        If False, raises FileNotFoundError if the directory's parent doesn't exist.
    """
    path = Path(path)
    old_umask = os.umask(umask)
    try:
        path.mkdir(exist_ok=exists_ok, parents=parents)
    finally:
        os.umask(old_umask)


def backoff_and_retry(
    backoff_seconds: int | float = 30,
    num_retries: int = 3,
    log_function: Callable[[str], None] = warnings.warn,
) -> Callable[[Callable[P, T]], Callable[P, T]]:
    """Adds a retry handler to the decorated function.

    Parameters
    ----------
    backoff_seconds
        Number of seconds to wait until retrying the operation.
    num_retries
        Number of times to retry before failing.
    log_function
        A callable used to emit retry messages.  Defaults to emitting
        a standard library warning.

    Returns
    -------
        A function that retries the decorated function a specified number of times.
    """

    def _wrap(func: Callable[P, T]) -> Callable[P, T]:
        @functools.wraps(func)
        def _wrapped(*args: P.args, **kwargs: P.kwargs) -> T:
            retries = num_retries
            while retries:
                try:
                    result = func(*args, **kwargs)
                    break
                except Exception as e:
                    log_function(
                        f"Error trying to write results, retries remaining {retries}"
                    )
                    time.sleep(backoff_seconds)
                    retries -= 1
                    if not retries:
                        log_function(f"Retries exhausted.")
                        raise e
            return result

        return _wrapped

    return _wrap



"""
==========================
vipin Performance Counters
==========================

Structs for counting performance metrics.

"""
from __future__ import annotations

import json
from time import time
from typing import TYPE_CHECKING

import psutil

if TYPE_CHECKING:
    from psutil import scpufreq, scpustats, sdiskio, snetio  # type: ignore[attr-defined]


class CounterSnapshot:
    def __init__(
        self,
        cpu: scpustats | None = None,
        disk: sdiskio | None = None,
        freq: scpufreq | None = None,
        net: snetio | None = None,
        timestamp: float | None = None,
    ) -> None:
        self.cpu = psutil.cpu_stats() if cpu is None else cpu
        self.disk = (
            psutil.disk_io_counters(perdisk=False, nowrap=True) if disk is None else disk
        )
        self.freq = psutil.cpu_freq(percpu=False) if freq is None else freq
        self.net = psutil.net_io_counters(pernic=False, nowrap=True) if net is None else net
        self.timestamp = time() if timestamp is None else timestamp

    def to_dict(self) -> dict[str, dict[str, int | float] | float]:
        c_dict: dict[str, dict[str, int | float] | float] = {}
        # Handle potential None values from psutil calls
        if self.cpu is not None:
            c_dict["cpu"] = dict(self.cpu._asdict())
        if self.freq is not None:
            c_dict["freq"] = dict(self.freq._asdict())
        if self.disk is not None:
            c_dict["disk"] = dict(self.disk._asdict())
        if self.net is not None:
            c_dict["net"] = dict(self.net._asdict())
        c_dict["time"] = self.timestamp
        return c_dict

    def __sub__(self, other: "CounterSnapshot") -> "CounterSnapshot":
        # Handle potential None values and use Any for psutil types
        cpu = None
        if self.cpu is not None and other.cpu is not None:
            cpu = type(other.cpu)(
                *tuple(self.cpu[i] - other.cpu[i] for i in range(len(other.cpu)))
            )

        disk = None
        if self.disk is not None and other.disk is not None:
            disk = type(other.disk)(
                *tuple(self.disk[i] - other.disk[i] for i in range(len(other.disk)))
            )

        freq = None
        if self.freq is not None and other.freq is not None:
            freq = type(other.freq)(
                *tuple((self.freq[i] + other.freq[i]) / 2 for i in range(len(other.freq)))
            )

        net = None
        if self.net is not None and other.net is not None:
            net = type(other.net)(
                *tuple(self.net[i] - other.net[i] for i in range(len(other.net)))
            )

        timestamp = self.timestamp - other.timestamp
        return CounterSnapshot(cpu, disk, freq, net, timestamp)

    def __repr__(self) -> str:
        return json.dumps(self.to_dict())



from pathlib import Path
from unittest.mock import patch

import pytest
from pytest_mock import MockerFixture

from vivarium_cluster_tools.psimulate.jobs import JobParameters
from vivarium_cluster_tools.psimulate.worker.vivarium_work_horse import work_horse


@pytest.mark.parametrize("log_level", [0, 1, 2])
def test_logging_level(mocker: MockerFixture, tmp_path: Path, log_level: int) -> None:

    input_draw = 1
    random_seed = 2
    branch_configuration = {"branch_key": "branch_value"}
    job_parameters = JobParameters(
        model_specification="test_model_spec.yaml",
        branch_configuration=branch_configuration,
        input_draw=input_draw,
        random_seed=random_seed,
        results_path="~/tmp",
        backup_configuration={
            "backup_freq": 300,
            "backup_dir": tmp_path / "backups",
            "backup_metadata_path": tmp_path / "backups" / "backup_metadata.csv",
        },
        extras={
            "sim_verbosity": log_level,
        },
    )

    mocker.patch(
        "vivarium_cluster_tools.psimulate.worker.vivarium_work_horse.ENV_VARIABLES",
    )
    mocker.patch(
        "vivarium_cluster_tools.psimulate.worker.vivarium_work_horse.get_current_job",
    )
    mocker.patch("vivarium_cluster_tools.psimulate.worker.vivarium_work_horse.run_simulation")
    mocker.patch("vivarium_cluster_tools.psimulate.worker.vivarium_work_horse.remove_backups")
    mocker.patch(
        "vivarium_cluster_tools.psimulate.worker.vivarium_work_horse.get_sim_results"
    )
    mocker.patch(
        "vivarium_cluster_tools.psimulate.worker.vivarium_work_horse.format_and_record_details"
    )
    with patch(
        "vivarium_cluster_tools.psimulate.worker.vivarium_work_horse.ParallelSimulationContext"
    ) as mock_parallel_sim:
        work_horse(job_parameters.to_dict())
        mock_parallel_sim.assert_called_once_with(
            job_parameters.model_specification,
            configuration=job_parameters.sim_config,
            logging_verbosity=log_level,
        )



"""
======================
vivarium_cluster_tools
======================

Tools for working with :mod:`vivarium` on compute clusters.

"""
from vivarium_cluster_tools._version import __version__
from vivarium_cluster_tools.utilities import get_cluster_name, mkdir



"""
================
psimulate Runner
================

The main process loop for `psimulate` runs.

"""

import os
import shutil
from collections import defaultdict
from pathlib import Path
from time import sleep, time
from typing import Any

import pandas as pd
from loguru import logger
from vivarium.framework.utilities import collapse_nested_dict

from vivarium_cluster_tools import logs
from vivarium_cluster_tools.psimulate import (
    COMMANDS,
    branches,
    cluster,
    jobs,
    model_specification,
    paths,
    pip_env,
    redis_dbs,
)
from vivarium_cluster_tools.psimulate import results as psim_results
from vivarium_cluster_tools.psimulate import worker
from vivarium_cluster_tools.psimulate.paths import OutputPaths
from vivarium_cluster_tools.psimulate.performance_logger import (
    append_perf_data_to_central_logs,
)
from vivarium_cluster_tools.vipin.perf_report import PerformanceSummary, report_performance


def process_job_results(
    registry_manager: redis_dbs.RegistryManager,
    existing_metadata: pd.DataFrame,
    output_paths: OutputPaths,
    batch_size: int,
    output_file_size: int,
    no_batch: bool,
) -> dict[str, int | float]:
    unwritten_metadata = []
    unwritten_results = []
    batch_size = 0 if no_batch else batch_size
    status: dict[str, int | float] = defaultdict(int)

    # Initialize output file map from any existing results (for restart support)
    output_file_map = psim_results.OutputFileMap.from_existing_results(
        output_paths.results_dir, output_file_size
    )

    logger.info("Entering main processing loop.")
    start_time = time()
    try:
        while registry_manager.jobs_to_finish:
            sleep(5)
            for metadata, results in registry_manager.get_results():
                unwritten_metadata.append(metadata)
                unwritten_results.append(results)

            if len(unwritten_results) > batch_size:
                (
                    existing_metadata,
                    unwritten_metadata,
                    unwritten_results,
                ) = psim_results.write_results_batch(
                    output_paths,
                    existing_metadata,
                    unwritten_metadata,
                    unwritten_results,
                    batch_size,
                    output_file_map,
                    output_file_size,
                )

            status = registry_manager.update_and_report()
            logger.info(f"Unwritten results: {len(unwritten_results)}")
            logger.info(f"Elapsed time: {(time() - start_time)/60:.1f} minutes.")
    finally:
        # Flush all remaining results
        while unwritten_results:
            (
                existing_metadata,
                unwritten_metadata,
                unwritten_results,
            ) = psim_results.write_results_batch(
                output_paths,
                existing_metadata,
                unwritten_metadata,
                unwritten_results,
                batch_size=len(unwritten_results),
                output_file_map=output_file_map,
                output_file_size=output_file_size,
            )
            logger.info(f"Unwritten results: {len(unwritten_results)}")
            logger.info(f"Elapsed time: {(time() - start_time) / 60:.1f} minutes.")

    return status


def load_existing_output_metadata(metadata_path: Path, restart: bool) -> pd.DataFrame:
    try:
        existing_output_metadata = pd.read_csv(metadata_path)
    except FileNotFoundError:
        existing_output_metadata = pd.DataFrame()
    assert (
        existing_output_metadata.empty or restart
    ), "How do you have existing outputs on an initial run?"
    return existing_output_metadata


def report_initial_status(
    num_jobs_completed: int, finished_sim_metadata: pd.DataFrame, total_num_jobs: int
) -> None:
    if num_jobs_completed:
        logger.info(
            f"{num_jobs_completed} of {total_num_jobs} jobs completed in previous run."
        )
    extra_jobs_completed = num_jobs_completed - len(finished_sim_metadata)
    # NOTE: there can never be more rows in `finished_sim_metadata` than `num_jobs_completed`
    # because `num_jobs_completed` was calculated by comparing the keyspace to `finished_sim_metadata`.
    if extra_jobs_completed:
        raise RuntimeError(
            f"There are {extra_jobs_completed} jobs from the previous run which would not have been created "
            "with the configuration saved with that run. That either means that code "
            "has changed between then and now or that the outputs or configuration data "
            "have been modified."
        )


def try_run_vipin(output_paths: OutputPaths) -> None:
    log_path = output_paths.worker_logging_root
    try:
        perf_df = report_performance(
            input_directory=log_path, output_directory=log_path, output_hdf=False, verbose=1
        )
    except Exception as e:
        logger.warning(f"Performance reporting failed with: {e}")
        return

    try:
        if perf_df is not None and len(perf_df) > 0:
            append_perf_data_to_central_logs(perf_df, output_paths)
    except Exception as e:
        logger.warning(f"Appending performance data to central logs failed with: {e}")


def write_backup_metadata(
    backup_metadata_path: Path, parameters_by_job: dict[str, dict[str, Any]]
) -> None:
    lookup_table = []
    for job_id, params in parameters_by_job.items():
        job_dict = {
            "input_draw": params["input_draw"],
            "random_seed": params["random_seed"],
            "job_id": job_id,
        }
        branch_config = collapse_nested_dict(params["branch_configuration"])
        for k, v in branch_config:
            job_dict[k] = v
        lookup_table.append(job_dict)

    df = pd.DataFrame(lookup_table)
    df.to_csv(
        backup_metadata_path,
        index=False,
        mode="a",
        header=not os.path.exists(backup_metadata_path),
    )


def main(
    command: str,
    input_paths: paths.InputPaths,
    native_specification: cluster.NativeSpecification,
    max_workers: int | None,
    redis_processes: int,
    batch_size: int,
    output_file_size: int,
    no_batch: bool,
    backup_freq: int | None,
    extra_args: dict[str, Any],
) -> None:
    logger.info("Validating cluster environment.")
    cluster.validate_cluster_environment()

    # Generate programmatic representation of the output directory structure
    output_paths = paths.OutputPaths.from_entry_point_args(
        command=command,
        input_artifact_path=input_paths.artifact,
        result_directory=input_paths.result_directory,
        input_model_spec_path=input_paths.model_specification,
    )
    logger.info("Setting up output directory and all subdirectories.")
    output_paths.touch()

    logger.info("Setting up logging to files.")
    # Start sending logs to a file now that it exists.
    logs.configure_main_process_logging_to_file(output_paths.logging_root)
    logger.info("Validating programming environment.")
    # Either write a requirements.txt with the current environment
    # or verify the current environment matches the prior environment
    # used when doing a restart.
    pip_env.validate(output_paths.environment_file)

    logger.info(
        "Parsing input arguments into model specification and branches and writing to disk."
    )
    # Parse the branches configuration into a parameter space
    # and a flat representation of all parameters to be run.
    keyspace = branches.Keyspace.from_entry_point_args(
        input_branch_configuration_path=input_paths.branch_configuration,
        keyspace_path=output_paths.keyspace,
        branches_path=output_paths.branches,
        extras=extra_args,
    )
    # Throw that into our output directory. The keyspace output is
    # a cartesian product representation of the parameter space and
    # branches is a flat representation with the product expanded out.
    keyspace.persist(output_paths.keyspace, output_paths.branches)

    # Parse the model specification and resolve the artifact path
    # and then write to the output directory.
    model_spec = model_specification.parse(
        command=command,
        input_model_specification_path=input_paths.model_specification,
        artifact_path=input_paths.artifact,
        model_specification_path=output_paths.model_specification,
        results_root=output_paths.root,
        keyspace=keyspace,
    )
    model_specification.persist(model_spec, output_paths.model_specification)

    logger.info("Loading existing outputs if present.")
    # Load in any existing partial outputs if present.
    finished_sim_metadata = load_existing_output_metadata(
        metadata_path=output_paths.finished_sim_metadata,
        restart=command in [COMMANDS.restart, COMMANDS.expand],
    )

    logger.info("Parsing arguments into worker job parameters.")
    # Translate the keyspace into the list of jobs to actually run
    # after accounting for any partially present results.
    job_parameters, num_jobs_completed = jobs.build_job_list(
        command=command,
        model_specification_path=output_paths.model_specification,
        output_root=output_paths.root,
        keyspace=keyspace,
        finished_sim_metadata=finished_sim_metadata,
        backup_freq=backup_freq,
        backup_dir=output_paths.backup_dir,
        backup_metadata_path=output_paths.backup_metadata_path,
        extras=extra_args,
    )
    # Let the user know if something is fishy at this point.
    total_num_jobs = len(keyspace)
    report_initial_status(num_jobs_completed, finished_sim_metadata, total_num_jobs)
    if len(job_parameters) == 0:
        logger.info("No jobs to run, exiting.")
        return
    else:
        logger.info(f"Found {len(job_parameters)} jobs to run.")

    num_workers = len(job_parameters)
    if max_workers:
        num_workers = min(max_workers, num_workers)

    logger.info("Spinning up Redis DBs and connecting to main process.")
    # Spin up the job & result dbs and get back (hostname, port) pairs for all the dbs.
    redis_ports = redis_dbs.launch(
        num_processes=redis_processes,
        num_workers=num_workers,
        redis_logging_root=output_paths.logging_root,
    )
    # Spin up a unified interface to all the redis databases
    registry_manager = redis_dbs.RegistryManager(redis_ports, num_workers, num_jobs_completed)
    logger.info("Enqueuing jobs on Redis queues.")
    # Distribute all the remaining jobs across the job queues
    # in the redis databases.
    registry_manager.enqueue(
        jobs=job_parameters, workhorse_import_path=worker.WORK_HORSE_PATHS[command]
    )

    if backup_freq is not None:
        write_backup_metadata(
            backup_metadata_path=output_paths.backup_metadata_path,
            parameters_by_job=registry_manager.get_params_by_job(),
        )
    # Generate a worker template that chooses a redis DB at random to connect to.
    # This should (approximately) evenly distribute the workers over the work.
    redis_urls = [f"redis://{hostname}:{port}" for hostname, port in redis_ports]
    worker_template = (
        f"import random\nredis_urls = {redis_urls}\nREDIS_URL = random.choice(redis_urls)\n\n"
    )
    # Dump the worker config to a file we can pass to the workers on startup.
    output_paths.worker_settings.write_text(worker_template)
    worker_launch_script = worker.build_launch_script(
        worker_settings_file=output_paths.worker_settings,
        worker_log_directory=output_paths.worker_logging_root,
    )
    logger.info(f"Submitting redis workers to the cluster.")
    # Start an rq worker for every job using the cluster scheduler, up to an
    # optional user-specified limit. The workers will start as soon as they get
    # scheduled and start looking for work. They run in burst mode which means
    # they shut down if they can't find anything to do. This means it's
    # critical that we put the jobs on the queue before the workers land,
    # otherwise they'll just show up and shut down.

    cluster.submit_worker_jobs(
        num_workers=num_workers,
        worker_launch_script=worker_launch_script,
        cluster_logging_root=output_paths.cluster_logging_root,
        native_specification=native_specification,
    )

    logger.info(
        "Entering monitoring and results processing loop. "
        f"Results will be written to {str(output_paths.root)}"
    )
    # Enter the main monitoring and processing loop, which will check on
    # all the queues periodically, report status updates, and gather
    # and write results when they are available.
    status = process_job_results(
        registry_manager=registry_manager,
        existing_metadata=finished_sim_metadata,
        output_paths=output_paths,
        batch_size=batch_size,
        output_file_size=output_file_size,
        no_batch=no_batch,
    )

    # Spit out a performance report for the workers.
    try_run_vipin(output_paths)

    # Emit warning if any jobs failed
    if status["failed"] > 0:
        logger.warning(
            f"*** NOTE: There {'was' if status['failed'] == 1 else 'were'} "
            f"{status['failed']} failed job{'' if status['failed'] == 1 else 's'}. ***"
        )
    else:
        logger.info(f"Removing sim backup directory {output_paths.backup_dir}")
        shutil.rmtree(output_paths.backup_dir)

    logger.info(
        f"{status['successful'] - num_jobs_completed} of {status['total']} jobs "
        f"completed successfully from this {command}.\n"
        f"({status['successful']} of {total_num_jobs} total jobs completed successfully overall)\n"
        f"Results written to: {str(output_paths.results_dir)}"
    )



"""
==================
Results Processing
==================

Tools for processing and writing results.

"""

import time
from collections import defaultdict
from pathlib import Path

import pandas as pd
import pyarrow.parquet as pq
from loguru import logger

from vivarium_cluster_tools import utilities as vct_utils
from vivarium_cluster_tools.psimulate.paths import OutputPaths

INITIAL_SAMPLE_ROWS = 1000


class OutputFileMap:
    """Tracks current output file number per metric during results writing.

    This is a mutable state mapping that gets updated in place as output files rotate.

    Attributes
    ----------
    results_dir
        Directory containing metric subdirectories with chunked output files.
    output_file_size
        Target size in bytes for each output file.
    metrics
        Dictionary mapping metric name to current file number. Defaults to 0 for new metrics.
    """

    def __init__(
        self,
        results_dir: Path,
        output_file_size: int,
        metrics: dict[str, int] | None = None,
    ) -> None:
        self.results_dir = results_dir
        self.output_file_size = output_file_size
        self.metrics: defaultdict[str, int] = defaultdict(int)
        self._bytes_per_row: dict[str, float] = {}
        if metrics is not None:
            self.metrics.update(metrics)

    @classmethod
    def from_existing_results(
        cls, results_dir: Path, output_file_size: int
    ) -> "OutputFileMap":
        """Create OutputFileMap by scanning existing chunked output files.

        Parameters
        ----------
        results_dir
            Directory containing metric subdirectories with chunked output files.
        output_file_size
            Target size in bytes for each output file.

        Returns
        -------
            OutputFileMap initialized with current output file numbers per metric.
        """
        metrics: dict[str, int] = {}

        if not results_dir.exists():
            return cls(results_dir, output_file_size, metrics)

        for metric_dir in results_dir.iterdir():
            if not metric_dir.is_dir():
                continue

            existing_output_files = sorted(metric_dir.glob("[0-9]*.parquet"))
            if existing_output_files:
                metrics[metric_dir.name] = int(existing_output_files[-1].stem)

        return cls(results_dir, output_file_size, metrics)

    def get_path(self, metric: str) -> Path:
        """Get full path to current chunk file for a metric."""
        output_file_num = self.metrics[metric]
        return self.results_dir / metric / f"{output_file_num:04d}.parquet"

    def __getitem__(self, metric: str) -> int:
        """Get current file number for a metric."""
        return self.metrics[metric]

    def __setitem__(self, metric: str, output_file_num: int) -> None:
        """Set file number for a metric."""
        self.metrics[metric] = output_file_num

    def __contains__(self, metric: str) -> bool:
        """Check if metric is tracked."""
        return metric in self.metrics

    def bytes_per_row(self, metric: str) -> float:
        """Get estimated bytes per row for a metric.

        While still filling the initial output file (0000.parquet), we dynamically
        recalculate from actual file size. Once that file is full (i.e., we've
        moved to file 0001 or higher), the estimate is locked in.

        Parameters
        ----------
        metric
            The metric name.

        Returns
        -------
            Estimated bytes per row.
        """
        # If we've already locked in an estimate, use it
        if metric in self._bytes_per_row:
            return self._bytes_per_row[metric]

        output_file_0_path = self.results_dir / metric / "0000.parquet"

        # Calculate dynamically from data in file 0
        if output_file_0_path.exists():
            rows: int = pq.read_metadata(output_file_0_path).num_rows
            if rows > 0:
                bytes_per_row = output_file_0_path.stat().st_size / rows
                if self.metrics[metric] > 0:
                    # If we've moved past file 0, lock in the estimate
                    self._bytes_per_row[metric] = bytes_per_row
                return bytes_per_row

        # No data yet - use initial estimate
        return self.output_file_size / INITIAL_SAMPLE_ROWS


def write_results_batch(
    output_paths: OutputPaths,
    existing_metadata: pd.DataFrame,
    unwritten_metadata: list[pd.DataFrame],
    unwritten_results: list[dict[str, pd.DataFrame]],
    batch_size: int,
    output_file_map: OutputFileMap,
    output_file_size: int,
) -> tuple[pd.DataFrame, list[pd.DataFrame], list[dict[str, pd.DataFrame]]]:
    """Write batch of results and finished simulation metadata to disk.

    Results are written to chunked parquet files within per-metric directories
    (e.g., ``results_dir/metric_name/0000.parquet``). When a file
    exceeds ``output_file_size`` bytes, a new file is started.

    To read all results for a metric, use ``pd.read_parquet(results_dir / metric_name)``,
    which automatically combines all parquet files in the directory.

    Parameters
    ----------
    output_paths
        Container class with output filepaths as attributes.
    existing_metadata
        Metadata for finished simulations that has already been written to disk.
    unwritten_metadata
        Metadata for finished simulations that has not yet been written to disk.
    unwritten_results
        Results for finished simulations that have not yet been written to disk.
    batch_size
        Number of results to write in this batch.
    output_file_map
        Dictionary mapping metric names to current file numbers. Updated in place.
    output_file_size
        Maximum file size in bytes. When exceeded, a new file is started.

    Returns
    -------
        A tuple of:
        - The updated metadata that was written to disk as part of this batch.
        - The unwritten metadata remaining to write in subsequent batches.
        - The unwritten results remaining to write in subsequent batches.
    """
    logger.info(f"Writing batch of {batch_size} results.")
    new_metadata_to_write, unwritten_metadata = (
        unwritten_metadata[:batch_size],
        unwritten_metadata[batch_size:],
    )
    new_results_to_write, unwritten_results = (
        unwritten_results[:batch_size],
        unwritten_results[batch_size:],
    )
    metadata_to_write = _concat_metadata(existing_metadata, new_metadata_to_write)
    results_to_write = _combine_batch_results(new_results_to_write)

    start = time.time()
    # Write results to files per metric
    for metric, new_df in results_to_write.items():
        (output_paths.results_dir / metric).mkdir(exist_ok=True)
        _write_output_file_per_metric(
            new_data=new_df,
            output_file_map=output_file_map,
            metric=metric,
            output_file_size=output_file_size,
        )
    # Metadata is small enough to overwrite entirely each time
    _safe_write(metadata_to_write, output_paths.finished_sim_metadata)
    end = time.time()
    logger.info(f"Updated results in {end - start:.4f}s.")
    return metadata_to_write, unwritten_metadata, unwritten_results


def _write_output_file_per_metric(
    new_data: pd.DataFrame,
    output_file_map: OutputFileMap,
    metric: str,
    output_file_size: int,
) -> None:
    """Write new data to chunk file(s), splitting across files if needed.

    Data may be split across multiple chunked output files to respect output_file_size limits.
    Each file will be approximately output_file_size bytes or smaller.
    """
    remaining = new_data.reset_index(drop=True)

    while not remaining.empty:
        output_file_path = output_file_map.get_path(metric)
        output_file_exists = output_file_path.exists()

        if output_file_exists:
            current_size = output_file_path.stat().st_size
            remaining_space = max(0, output_file_size - current_size)
        else:
            remaining_space = output_file_size

        rows_that_fit = int(remaining_space / output_file_map.bytes_per_row(metric))
        if rows_that_fit < 1:
            single_row_too_large = output_file_map.bytes_per_row(metric) > output_file_size
            file_empty = not output_file_exists or remaining_space == output_file_size

            if single_row_too_large and file_empty:
                # If the file size is smaller than estimated bytes per row,
                # we have no choice but to write one row per output file
                logger.warning(
                    f"Estimated bytes per row for metric '{metric}' "
                    f"({output_file_map.bytes_per_row(metric):.2f} bytes) "
                    f"exceeds output file size ({output_file_size} bytes). "
                    f"Writing one row per output file."
                )
                rows_that_fit = 1
            else:
                # No space left in current output file, rotate to next chunk
                output_file_map[metric] += 1
                continue

        to_write = remaining.iloc[:rows_that_fit].reset_index(drop=True)
        remaining = remaining.iloc[rows_that_fit:].reset_index(drop=True)

        # Combine with existing data if present
        if output_file_exists:
            combined = _concat_preserve_types(
                [pd.read_parquet(output_file_path).reset_index(drop=True), to_write]
            )
        else:
            combined = to_write

        _safe_write(combined, output_file_path)


def _concat_metadata(
    old_metadata: pd.DataFrame, new_metadata: list[pd.DataFrame]
) -> pd.DataFrame:
    """Concatenate the new metadata to the old."""
    start = time.time()

    to_concat_old = [old_metadata.reset_index(drop=True)] if not old_metadata.empty else []
    to_concat = to_concat_old + [df.reset_index(drop=True) for df in new_metadata]

    updated = _concat_preserve_types(to_concat)

    end = time.time()
    logger.info(f"Concatenated {len(new_metadata)} metadata in {end - start:.2f}s.")
    return updated


def _combine_batch_results(
    new_results: list[dict[str, pd.DataFrame]],
) -> dict[str, pd.DataFrame]:
    """Concatenate only the new batch results (not existing results on disk).

    This function combines the results from the current batch into a single
    DataFrame per metric, without loading any existing results from disk.
    """
    start = time.time()
    results = {}
    metrics = {key for new in new_results for key in new.keys()}
    for metric in metrics:
        to_concat = [
            df.reset_index(drop=True)
            for result in new_results
            for met, df in result.items()
            if met == metric
        ]
        results[metric] = _concat_preserve_types(to_concat)

    end = time.time()
    logger.info(f"Concatenated {len(new_results)} results in {end - start:.2f}s.")
    return results


def _concat_preserve_types(df_list: list[pd.DataFrame]) -> pd.DataFrame:
    """Concatenate datasets and preserve all ``numpy`` dtypes.

    This does not preserve any pandas-specific dtypes, e.g. categories become
    objects). We assume that all dataframes in the list have identical columns
    and dtypes to the first.
    """
    dtype_mapping = {col: df_list[0][col].dtype for col in df_list[0].columns}
    for df in df_list:
        for col, dtype in dtype_mapping.items():
            df[col] = df[col].astype(dtype)
    return pd.concat(df_list, axis=0).reset_index(drop=True)


@vct_utils.backoff_and_retry(backoff_seconds=30, num_retries=3, log_function=logger.warning)
def _safe_write(df: pd.DataFrame, output_path: Path) -> None:
    """Write a dataframe to disk, retrying if there are any issues.

    Parameters
    ----------
    df
        DataFrame to write to disk.
    output_path
        Path to write the DataFrame to.

    Raises
    ------
    NotImplementedError
        If the file extension is not supported.
    """
    # Writing to some file types over and over balloons the file size so
    # write to new file and move it over to avoid
    temp_output_path = output_path.with_name(
        output_path.stem + "_update" + output_path.suffix
    )
    if output_path.suffix == ".parquet":
        df.to_parquet(temp_output_path)
    elif output_path.suffix == ".csv":
        df.to_csv(temp_output_path, index=False)
    else:
        raise NotImplementedError(f"Writing to {output_path.suffix} is not supported.")
    temp_output_path.replace(output_path)



"""
==============================
Results Processing CLI Options
==============================

Command line options for configuring results handling in psimulate runs.

"""

import click

from vivarium_cluster_tools.cli_tools import MINUTES_OR_NONE

DEFAULT_BATCH_SIZE = 200
DEFAULT_OUTPUT_FILE_SIZE_MB = 500.0

with_batch_size = click.option(
    "--batch-size",
    type=int,
    default=DEFAULT_BATCH_SIZE,
    show_default=True,
    help="Number of simulation jobs to accumulate before writing results to disk.",
)
with_output_file_size = click.option(
    "--output-file-size",
    "output_file_size",
    type=float,
    default=DEFAULT_OUTPUT_FILE_SIZE_MB,
    show_default=True,
    help="Maximum file size in MB per result file (within each type of result). When exceeded, a new file is started.",
    callback=lambda ctx, param, value: int(value * 1024 * 1024),  # Convert MB to bytes
)
with_no_batch = click.option(
    "--no-batch",
    is_flag=True,
    help="Write results immediately as they come in (batch_size=0). Files are still split by size per --output_file_size.",
)
with_no_cleanup = click.option(
    "--no-cleanup",
    is_flag=True,
    hidden=True,
    help="Hidden developer option, if flagged, don't automatically "
    "cleanup results directory on failure.",
)
backup_freq = click.option(
    "--backup-freq",
    type=MINUTES_OR_NONE,
    default="30",
    show_default=True,
    help="Interval in minutes between saving backups. Set to 'None' or 'none' to disable backups.",
)
